@@ -75,7 +75,6 @@ MARIA_HA *_ma_test_if_reopen(const char *filename)
   SYNOPSIS
     maria_clone_internal()
     share	Share of already open table
-    mode	Mode of table (O_RDONLY | O_RDWR)
     data_file   Filedescriptor of data file to use < 0 if one should open
 	        open it.
     internal_table <> 0 if this is an internal temporary table
@@ -86,7 +85,7 @@ MARIA_HA *_ma_test_if_reopen(const char *filename)
 */
 
 static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
-                                      int mode, File data_file,
+                                      File data_file,
                                       uint internal_table,
                                       struct ms3_st *s3)
 {
@@ -100,11 +99,6 @@ static MARIA_HA *maria_clone_internal(MARIA_SHARE *share,
   errpos= 0;
   bzero((uchar*) &info,sizeof(info));
 
-  if (mode == O_RDWR && share->mode == O_RDONLY)
-  {
-    my_errno=EACCES;				/* Can't open in write mode */
-    goto err;
-  }
   if (data_file >= 0)
     info.dfile.file= data_file;
   else if (_ma_open_datafile(&info, share))
@@ -261,7 +255,8 @@ err:
 MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
                      S3_INFO *s3)
 {
-  int open_mode= 0,save_errno;
+  int save_errno;
+  int open_mode= mode;
   uint i,j,len,errpos,head_length,base_pos,keys, realpath_err,
     key_parts,base_key_parts,unique_key_parts,fulltext_keys,uniques;
   uint internal_table= MY_TEST(open_flags & HA_OPEN_INTERNAL_TABLE);
@@ -293,6 +288,11 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
 
 #ifndef WITH_S3_STORAGE_ENGINE
   DBUG_ASSERT(!s3);
+# ifdef _MSC_VER
+  __assume(!s3);
+# else
+  if (s3) __builtin_unreachable();
+# endif
 #else
   if (!s3)
 #endif /* WITH_S3_STORAGE_ENGINE */
@@ -345,14 +345,25 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
                         goto err;
                       });
       DEBUG_SYNC_C("mi_open_kfile");
+
+      /*
+        We first try to open the file on read-write mode to ensure
+        that the table is usable for future read and write queries in
+        MariaDB.  Only if the read-write mode fails we try to readonly.
+      */
+      if (!(open_flags & HA_OPEN_FORCE_MODE))
+        open_mode= O_RDWR;
+
       if ((kfile=mysql_file_open(key_file_kfile, name_buff,
-                                 (open_mode=O_RDWR) | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
+                                 open_mode | O_SHARE |
+                                 O_NOFOLLOW | O_CLOEXEC,
                                  MYF(common_flag | MY_NOSYMLINKS))) < 0)
       {
-        if ((errno != EROFS && errno != EACCES) ||
+        if ((errno != EROFS && errno != EACCES) || open_mode == O_RDONLY ||
             mode != O_RDONLY ||
             (kfile=mysql_file_open(key_file_kfile, name_buff,
-                                   (open_mode=O_RDONLY) | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
+                                   (open_mode=O_RDONLY) | O_SHARE |
+                                   O_NOFOLLOW | O_CLOEXEC,
                                    MYF(common_flag | MY_NOSYMLINKS))) < 0)
           goto err;
       }
@@ -367,7 +378,6 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
 #ifdef WITH_S3_STORAGE_ENGINE
     else
     {
-      open_mode= mode;
       errpos= 1;
       if (s3f.set_database_and_table_from_path(s3, name_buff))
       {
@@ -400,7 +410,9 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
     }
 #endif /* WITH_S3_STORAGE_ENGINE */
 
-    share->mode=open_mode;
+    share->index_mode= share->data_mode= open_mode;
+    if (open_flags & HA_OPEN_DATA_READONLY)
+      share->data_mode= O_RDONLY;
     if (memcmp(share->state.header.file_version, maria_file_magic, 4))
     {
       DBUG_PRINT("error",("Wrong header in %s",name_buff));
@@ -447,7 +459,9 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
           my_errno= HA_WRONG_CREATE_OPTION;
           goto err;
         }
-        share->mode|= O_NOFOLLOW; /* all symlinks are resolved by realpath() */
+        /* all symlinks are resolved by realpath() */
+        share->index_mode|= O_NOFOLLOW;
+        share->data_mode|= O_NOFOLLOW;
       }
     }
     else
@@ -760,19 +774,14 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
 	  else if (pos->type == HA_KEYTYPE_BINARY)
 	    pos->charset= &my_charset_bin;
 	}
-	if (keyinfo->flag & HA_SPATIAL)
+	if (keyinfo->key_alg == HA_KEY_ALG_RTREE)
 	{
-#ifdef HAVE_SPATIAL
 	  uint sp_segs=SPDIMS*2;
 	  keyinfo->seg=pos-sp_segs;
 	  keyinfo->keysegs--;
           versioning= 0;
-#else
-	  my_errno=HA_ERR_UNSUPPORTED;
-	  goto err;
-#endif
 	}
-        else if (keyinfo->flag & HA_FULLTEXT)
+        else if (keyinfo->key_alg == HA_KEY_ALG_FULLTEXT)
 	{
           versioning= 0;
           DBUG_ASSERT(fulltext_keys);
@@ -796,6 +805,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
             memcpy(&share->ft2_keyinfo, keyinfo, sizeof(MARIA_KEYDEF));
             share->ft2_keyinfo.keysegs=1;
             share->ft2_keyinfo.flag=0;
+            share->ft2_keyinfo.key_alg=HA_KEY_ALG_BTREE;
             share->ft2_keyinfo.keylength=
             share->ft2_keyinfo.minlength=
             share->ft2_keyinfo.maxlength=HA_FT_WLEN+share->base.rec_reflength;
@@ -1172,7 +1182,7 @@ MARIA_HA *maria_open(const char *name, int mode, uint open_flags,
     s3f.free(&index_header);
 #endif /* WITH_S3_STORAGE_ENGINE */
 
-  if (!(m_info= maria_clone_internal(share, mode, data_file,
+  if (!(m_info= maria_clone_internal(share, data_file,
                                      internal_table, s3_client)))
     goto err;
 
@@ -1401,19 +1411,15 @@ static void setup_key_functions(register MARIA_KEYDEF *keyinfo)
 {
   if (keyinfo->key_alg == HA_KEY_ALG_RTREE)
   {
-#ifdef HAVE_RTREE_KEYS
     keyinfo->ck_insert = maria_rtree_insert;
     keyinfo->ck_delete = maria_rtree_delete;
-#else
-    DBUG_ASSERT(0); /* maria_open should check it never happens */
-#endif
   }
   else
   {
     keyinfo->ck_insert = _ma_ck_write;
     keyinfo->ck_delete = _ma_ck_delete;
   }
-  if (keyinfo->flag & HA_SPATIAL)
+  if (keyinfo->key_alg == HA_KEY_ALG_RTREE)
     keyinfo->make_key= _ma_sp_make_key;
   else
     keyinfo->make_key= _ma_make_key;
@@ -1468,12 +1474,16 @@ static void setup_key_functions(register MARIA_KEYDEF *keyinfo)
   /* set keyinfo->write_comp_flag */
   if (keyinfo->flag & HA_SORT_ALLOWS_SAME)
     keyinfo->write_comp_flag=SEARCH_BIGGER; /* Put after same key */
-  else if (keyinfo->flag & ( HA_NOSAME | HA_FULLTEXT))
+  else if (keyinfo->flag & HA_NOSAME)
   {
     keyinfo->write_comp_flag= SEARCH_FIND | SEARCH_UPDATE; /* No duplicates */
     if (keyinfo->flag & HA_NULL_ARE_EQUAL)
       keyinfo->write_comp_flag|= SEARCH_NULL_ARE_EQUAL;
   }
+  else if (keyinfo->key_alg == HA_KEY_ALG_FULLTEXT)
+  {
+    keyinfo->write_comp_flag= SEARCH_FIND | SEARCH_UPDATE;
+  }
   else
     keyinfo->write_comp_flag= SEARCH_SAME; /* Keys in rec-pos order */
   keyinfo->write_comp_flag|= SEARCH_INSERT;
@@ -1559,7 +1569,7 @@ uint _ma_state_info_write(MARIA_SHARE *share, uint pWrite)
      @retval 1      Error
 */
 
-/* Stack size 26376 from clang */
+/* MARIA_STATE_INFO_SIZE + MARIA_STATE_EXTRA_SIZE == 25559 */
 PRAGMA_DISABLE_CHECK_STACK_FRAME
 
 uint _ma_state_info_write_sub(File file, MARIA_STATE_INFO *state, uint pWrite)
@@ -2054,12 +2064,12 @@ void _ma_set_index_pagecache_callbacks(PAGECACHE_FILE *file,
 
 int _ma_open_datafile(MARIA_HA *info, MARIA_SHARE *share)
 {
-  myf flags= ((share->mode & O_NOFOLLOW) ? MY_NOSYMLINKS | MY_WME : MY_WME) |
-    share->malloc_flag;
+  myf flags= ((share->data_mode & O_NOFOLLOW) ?
+              MY_NOSYMLINKS | MY_WME : MY_WME) | share->malloc_flag;
   DEBUG_SYNC_C("mi_open_datafile");
   info->dfile.file= share->bitmap.file.file=
     mysql_file_open(key_file_dfile, share->data_file_name.str,
-                    share->mode | O_SHARE | O_CLOEXEC, flags);
+                    share->data_mode | O_SHARE | O_CLOEXEC, flags);
   return info->dfile.file >= 0 ? 0 : 1;
 }
 
@@ -2073,8 +2083,9 @@ int _ma_open_keyfile(MARIA_SHARE *share)
   mysql_mutex_lock(&share->intern_lock);
   share->kfile.file= mysql_file_open(key_file_kfile,
                                      share->unique_file_name.str,
-                                     share->mode | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
-                             MYF(MY_WME | MY_NOSYMLINKS));
+                                     share->index_mode | O_SHARE | O_NOFOLLOW |
+                                     O_CLOEXEC,
+                                     MYF(MY_WME | MY_NOSYMLINKS));
   mysql_mutex_unlock(&share->intern_lock);
   return (share->kfile.file < 0);
 }
@@ -2166,7 +2177,7 @@ int maria_indexes_are_disabled(MARIA_HA *info)
 
   /*
     No keys or all are enabled. keys is the number of keys. Left shifted
-    gives us only one bit set. When decreased by one, gives us all all bits
+    gives us only one bit set. When decreased by one, gives us all bits
     up to this one set and it gets unset.
   */
   if (!share->base.keys ||
